#include <QLabel>
#include <QLineEdit>
#include <QScrollBar>
#include "HexWidget.hxx"
#include "ui_HexWidget.h"

// --- internal typedefs ---

using QL = QLabel;
using QLE = QLineEdit;
using QTW = QTableWidget;
using QTWI = QTableWidgetItem;

// --- internal constants ---

const QChar Null('0');

// --- constructors and deconstructors ---

HexWidget::HexWidget(QWidget *parent)
: QWidget(parent), _ui(new Ui::HexWidget), _width(8)
{
    _ui->setupUi(this);

    setupActions();
}

HexWidget::~HexWidget()
{
    delete _ui;
}

// --- public methods ---

QByteArray HexWidget::hexData() const
{
    return _data;
}

void HexWidget::setHexData(const QByteArray& data)
{
    _data = data;

    slotRefreshViews();
}

// --- public slots ---

void HexWidget::slotRefreshViews()
{
    const qint32 max_chars = _data.count();
    const qint32 max_rows = max_chars / _width;
    QTW *dline = _ui->wid_decline;
    QTW *hex = _ui->wid_hex;
    QTW *asc = _ui->wid_asc;
    QTW *hline = _ui->wid_hexline;

    dline->setColumnCount(1);
    dline->setRowCount(max_rows);
    hex->setColumnCount(_width);
    hex->setRowCount(max_rows);
    asc->setColumnCount(_width);
    asc->setRowCount(max_rows);
    hline->setColumnCount(1);
    hline->setRowCount(max_rows);

    for (qint32 row = 0; row < max_rows; ++row)
    {
        QTWI *dl = new QTWI(QString("%1").arg(row * _width, 10, 10, Null));
        QTWI *hl = new QTWI((QString("%1").arg(row * _width, 8, 16, Null)).toUpper());

        dl->setFlags(dl->flags() ^ Qt::ItemIsEditable);
        hl->setFlags(dl->flags() ^ Qt::ItemIsEditable);

        dline->setItem(row, 0, dl);
        hline->setItem(row, 0, hl);

        for (qint32 col = 0; col < _width; ++col)
        {
            const qint32 pos = row * _width + col;

            if (pos < max_chars)
            {
                const unsigned char chr = _data.at(pos);
                QTWI *he = new QTWI(QString("%1").arg(chr, 2, 16, Null));
                QTWI *ae = new QTWI(QString(chr));

                he->setFlags(he->flags() ^ Qt::ItemIsEditable);
                ae->setFlags(ae->flags() ^ Qt::ItemIsEditable);

                hex->setItem(row, col, he);
                asc->setItem(row, col, ae);
            }
        }
    }

    dline->resizeColumnsToContents();
    dline->resizeRowsToContents();

    hex->resizeColumnsToContents();
    hex->resizeRowsToContents();

    asc->resizeColumnsToContents();
    asc->resizeRowsToContents();
    for (qint32 col = 1; col < _width; ++col)
        asc->setColumnWidth(col, asc->columnWidth(0));

    hline->resizeColumnsToContents();
    hline->resizeRowsToContents();
}

void HexWidget::slotSelectCellDLine(qint32 row, qint32 col)
{
    _ui->wid_hex->setCurrentCell(row, col);
    _ui->wid_asc->setCurrentCell(row, col);
    _ui->wid_hexline->setCurrentCell(row, 0);
}

void HexWidget::slotSelectCellHex(qint32 row, qint32 col)
{
    _ui->wid_decline->setCurrentCell(row, 0);
    _ui->wid_asc->setCurrentCell(row, col);
    _ui->wid_hexline->setCurrentCell(row, 0);
}

void HexWidget::slotSelectCellAsc(qint32 row, qint32 col)
{
    _ui->wid_decline->setCurrentCell(row, 0);
    _ui->wid_hex->setCurrentCell(row, col);
    _ui->wid_hexline->setCurrentCell(row, 0);
}

void HexWidget::slotSelectCellHLine(qint32 row, qint32 col)
{
    _ui->wid_decline->setCurrentCell(row, 0);
    _ui->wid_hex->setCurrentCell(row, col);
    _ui->wid_asc->setCurrentCell(row, col);
}

void HexWidget::slotScrollDLine(qint32 val)
{
    _ui->wid_hex->verticalScrollBar()->setValue(val);
    _ui->wid_asc->verticalScrollBar()->setValue(val);
    _ui->wid_hexline->verticalScrollBar()->setValue(val);
}

void HexWidget::slotScrollHex(qint32 val)
{
    _ui->wid_decline->verticalScrollBar()->setValue(val);
    _ui->wid_asc->verticalScrollBar()->setValue(val);
    _ui->wid_hexline->verticalScrollBar()->setValue(val);
}

void HexWidget::slotScrollAsc(qint32 val)
{
    _ui->wid_decline->verticalScrollBar()->setValue(val);
    _ui->wid_hex->verticalScrollBar()->setValue(val);
    _ui->wid_hexline->verticalScrollBar()->setValue(val);
}

void HexWidget::slotScrollHLine(qint32 val)
{
    _ui->wid_decline->verticalScrollBar()->setValue(val);
    _ui->wid_hex->verticalScrollBar()->setValue(val);
    _ui->wid_asc->verticalScrollBar()->setValue(val);
}

// --- protected methods ---

void HexWidget::changeEvent(QEvent *event)
{
    if (event->type() == QEvent::LanguageChange)
        _ui->retranslateUi(this);
    else
        QWidget::changeEvent(event);
}

void HexWidget::setupActions()
{
    connect(_ui->wid_decline, SIGNAL(cellClicked(qint32, qint32)),
            this, SLOT(slotSelectCellDLine(qint32, qint32)));
    connect(_ui->wid_hex, SIGNAL(cellClicked(qint32, qint32)),
            this, SLOT(slotSelectCellHex(qint32, qint32)));
    connect(_ui->wid_asc, SIGNAL(cellClicked(qint32, qint32)),
            this, SLOT(slotSelectCellAsc(qint32, qint32)));
    connect(_ui->wid_hexline, SIGNAL(cellClicked(qint32, qint32)),
            this, SLOT(slotSelectCellHLine(qint32, qint32)));

    connect(_ui->wid_decline->verticalScrollBar(), SIGNAL(valueChanged(qint32)),
            this, SLOT(slotScrollDLine(qint32)));
    connect(_ui->wid_hex->verticalScrollBar(), SIGNAL(valueChanged(qint32)),
            this, SLOT(slotScrollHex(qint32)));
    connect(_ui->wid_asc->verticalScrollBar(), SIGNAL(valueChanged(qint32)),
            this, SLOT(slotScrollAsc(qint32)));
    connect(_ui->wid_hexline->verticalScrollBar(), SIGNAL(valueChanged(qint32)),
            this, SLOT(slotScrollHLine(qint32)));
}
