#pragma once

#include <sstream>
#include "Vector.hxx"

template <typename MT>
class Ray {
    static_assert (std::is_integral<MT>::value || std::is_floating_point<MT>::value,
                   "ERROR: template parameter is not an integral or floating point type");
public:
    // --- constructors and deconstructors ---

    Ray()
    : _origin(0, 0, 0), _direction(0, 0, 0), _rgb(0, 0, 0), _min(5), _max(1000), _depth(10), _valid(false)
    {
    }

    Ray(const Vector3<MT>& org, const Vector3<MT>& dir, const Vector3<MT>& rgb, const MT& min, const MT& max,
        const size_t& depth, const bool valid)
    : _origin(org), _direction(dir), _rgb(rgb), _min(min), _max(max), _depth(depth), _valid(valid)
    {
    }

    Ray(const Ray& rhs)
    : _origin(rhs._origin), _direction(rhs._direction), _rgb(rhs._rgb), _min(rhs._min), _max(rhs._max),
      _depth(rhs._depth), _valid(rhs._valid)
    {
    }

    Ray(Ray&& rhs)
    : _origin(std::move(rhs._origin)), _direction(std::move(rhs._direction)), _rgb(std::move(rhs._rgb)),
      _min(std::move(rhs._min)), _max(std::move(rhs._max)), _depth(std::move(rhs._depth)), _valid(std::move(rhs._valid))
    {
    }

    virtual ~Ray()
    {
    }

    // --- default operators ---

    Ray& operator=(const Ray& rhs)
    {
        if (this != &rhs)
        {
            _origin = rhs._origin;
            _direction = rhs._direction;
            _rgb = rhs._rgb;
            _min = rhs._min;
            _max = rhs._max;
            _depth = rhs._depth;
            _valid = rhs._valid;
        }

        return *this;
    }

    Ray& operator=(Ray& rhs)
    {
        if (this != &rhs)
        {
            _origin = std::move(rhs._origin);
            _direction = std::move(rhs._direction);
            _rgb = std::move(rhs._rgb);
            _min = std::move(rhs._min);
            _max = std::move(rhs._max);
            _depth = std::move(rhs._depth);
            _valid = std::move(rhs._valid);
        }

        return *this;
    }

    bool operator==(const Ray& rhs) const
    {
        return _origin == rhs._origin && _direction == rhs._direction && _rgb == rhs._rgb && _min == rhs._min &&
               _max == rhs._max && _depth == rhs._depth && _valid == rhs._valid;
    }

    bool operator!=(const Ray& rhs) const
    {
        return !(*this == rhs);
    }

    // --- public methods ---

    void setOrigin(const Vector3<MT>& org)
    {
        _origin = org;
    }

    Vector3<MT> origin() const
    {
        return _origin;
    }

    void setDirection(const Vector3<MT>& dir)
    {
        _direction = dir.normalized();
    }

    Vector3<MT> direction() const
    {
        return _direction;
    }

    void setRgb(const Vector3<MT>& rgb)
    {
        _rgb = rgb;
    }

    Vector3<MT> rgb() const
    {
        return _rgb;
    }

    void setMin(const MT& min)
    {
        _min = min;
    }

    MT min() const
    {
        return _min;
    }

    void setMax(const MT& max)
    {
        _max = max;
    }

    MT max() const
    {
        return _max;
    }

    void setDepth(const size_t& depth)
    {
        _depth = depth;
    }

    size_t depth() const
    {
        return _depth;
    }

    void setValid(const bool valid)
    {
        _valid = valid;
    }

    bool valid() const
    {
        return _valid;
    }

    std::string string() const
    {
        std::stringstream result;

        result << "(" << _origin.string() << "), (" << _direction.string() << ")";

        return result.str();
    }

protected:
    // --- protected data structures ---

    Vector3<MT> _origin;
    Vector3<MT> _direction;
    Vector4<uint8_t> _rgb;  // color ARGB (0, 255 for now)
    MT _min;                // near plane value
    MT _max;                // far plane value
    size_t _depth;          // recursion depth
    bool _valid;
};
