#pragma once

/*
 *  Using Qt here would be nuts. Qt's vector components are no templates and I need some more control of that very
 *  important part of the engine code. So I made my own vector classes (also a bit more flexible) ...
 *  - Vector2 for GUI and 2D stuff
 *  - Vector3 for the common 3D stuff
 *  - Vector4 for the matrix and quaternion magic
 *  ~ Akiko ~
 */

#include <cmath>
#include <iostream>
#include <sstream>
#include <type_traits>

template <typename MT> class Vector2;
template <typename MT> class Vector3;
template <typename MT> class Vector4;

// ****************************************************************************************************************** //

template <typename MT>
class Vector2 {
    static_assert (std::is_integral<MT>::value || std::is_floating_point<MT>::value,
                   "ERROR: template parameter is not an integral or floating point type");
public:
    // --- Vector2: data structures ---

    alignas (16) union {
        struct {
            MT x;
            MT y;
        };
        MT xy[2];
    };

    // --- Vector2: constructors and deconstructors ---

    Vector2()
    : xy{0.0, 0.0}
    {
    }

    Vector2(const MT& val)
    : xy{val, val}
    {
    }

    Vector2(const MT& x, const MT& y)
    : xy{x, y}
    {
    }

    Vector2(const Vector3<MT>& vec)
    : xy{vec.x, vec.y}
    {
    }

    Vector2(const Vector4<MT>& vec)
    : xy{vec.x, vec.y}
    {
    }

    Vector2(const Vector2& rhs)
    : xy{rhs.x, rhs.y}
    {
    }

    Vector2(Vector2&& rhs)
    : xy{std::move(rhs.x), std::move(rhs.y)}
    {
    }

    virtual ~Vector2()
    {
    }

    // --- Vector2: default operators ---

    Vector2& operator=(const Vector2& rhs)
    {
        if (this != &rhs)
        {
            x = rhs.x;
            y = rhs.y;
        }

        return *this;
    }

    Vector2& operator=(Vector2&& rhs)
    {
        if (this != &rhs)
        {
            x = std::move(rhs.x);
            y = std::move(rhs.y);
        }

        return *this;
    }

    bool operator==(const Vector2& rhs)
    {
        return x == rhs.x && y == rhs.y;
    }

    bool operator!=(const Vector2& rhs)
    {
        return !(*this == rhs);
    }

    const MT operator[](const size_t& index) const
    {
        return xy[index];
    }

    MT& operator[](const size_t& index)
    {
        return xy[index];
    }

    Vector2& operator~()
    {
        x = -x;
        y = -y;

        return *this;
    }

    Vector2 operator-() const
    {
        return {-x, -y};
    }

    // --- Vector2: other operators ---

    Vector2& operator+=(const Vector2& rhs)
    {
        x += rhs.x;
        y += rhs.y;

        return *this;
    }

    template <typename T>
    Vector2& operator+=(const T& val)
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        x += val;
        y += val;

        return *this;
    }

    Vector2 operator+(const Vector2& rhs) const
    {
        return {x + rhs.x, y + rhs.y};
    }

    template <typename T>
    Vector2 operator+(const T& val) const
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        return {x + val, y + val};
    }

    // ---

    Vector2& operator-=(const Vector2& rhs)
    {
        x -= rhs.x;
        y -= rhs.y;

        return *this;
    }

    template <typename T>
    Vector2& operator-=(const T& val)
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        x -= val;
        y -= val;

        return *this;
    }

    Vector2 operator-(const Vector2& rhs) const
    {
        return {x - rhs.x, y - rhs.y};
    }

    template <typename T>
    Vector2 operator-(const T& val)
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        return {x - val, y - val};
    }

    // ---

    Vector2& operator*=(const Vector2& rhs)
    {
        x *= rhs.x;
        y *= rhs.y;

        return *this;
    }

    template <typename T>
    Vector2& operator*=(const T& val)
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        x *= val;
        y *= val;

        return *this;
    }

    Vector2 operator*(const Vector2& rhs) const
    {
        return {x * rhs.x, y * rhs.y};
    }

    template <typename T>
    Vector2 operator*(const T& val)
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        return {x * val, y * val};
    }

    // ---

    Vector2& operator/=(const Vector2& rhs)
    {
        x /= rhs.x;
        y /= rhs.y;

        return *this;
    }

    template <typename T>
    Vector2& operator/=(const T& val)
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        x /= val;
        y /= val;

        return *this;
    }

    Vector2 operator/(const Vector2& rhs) const
    {
        return {x / rhs.x, y / rhs.y};
    }

    template <typename T>
    Vector2 operator/(const T& val) const
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        return {x / val, y / val};
    }

    // --- Vector2: public methods ---

    MT distanceToLine(const Vector2& origin, const Vector2& direction) const
    {
        if (direction.isNull())
            return (*this - origin).length();

        return (*this - (origin + dotProduct(*this - origin, direction) * direction)).length();
    }

    MT distanceToLinePoints(const Vector2& point1, const Vector2& point2) const
    {
        const Vector2 direction = point2 - point1;

        if (direction.isNull())
            return (*this - point1).length();

        return (*this - (point1 + dotProduct(*this - point1, direction) * direction)).length();
    }

    MT distanceToPoint(const Vector2& point) const
    {
        return (*this - point).length();
    }

    MT dotProduct(const Vector2& vec) const
    {
        return x * vec.x + y * vec.y;
    }

    bool isNull() const
    {
        return x == 0.0 && y == 0.0;
    }

    MT length() const
    {
        return std::sqrt(x * x + y * y);
    }

    MT lengthSquared() const
    {
        return x * x + y * y;
    }

    void normalize()
    {
        *this /= std::sqrt(x * x + y * y);
    }

    Vector2 normalized() const
    {
        const MT len = std::sqrt(x * x + y * y);

        return {x / len, y / len};
    }

    std::string string() const
    {
        std::stringstream result;

        result << x << ", " << y;

        return result.str();
    }

    Vector3<MT> vector3() const
    {
        return {x, y, 0.0};
    }

    Vector4<MT> vector4() const
    {
        return {x, y, 0.0, 0.0};
    }

    // --- Vector2: static methods ---

    static MT toDotProduct(const Vector2& vec1, const Vector2& vec2)
    {
        return vec1.x * vec2.x + vec1.y * vec2.y;
    }

    static std::string toString(const Vector2& vec)
    {
        std::stringstream result;

        result << vec.x << ", " << vec.y;

        return result.str();
    }

    static Vector3<MT> toVector3(const Vector2& vec)
    {
        return {vec.x, vec.y, 0.0};
    }

    static Vector4<MT> toVector4(const Vector2& vec)
    {
        return {vec.x, vec.y, 0.0, 0.0};
    }
};

// --- Vector2: global operators ---

template <typename T>
bool operator==(const Vector2<T>& lhs, const Vector2<T>& rhs)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return lhs.x == rhs.x && lhs.y == rhs.y;
}

template <typename T>
bool operator!=(const Vector2<T> lhs, const Vector2<T>& rhs)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return !(lhs == rhs);
}

template <typename T>
std::ostream& operator<<(std::ostream& lhs, const Vector2<T>& rhs)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    lhs << rhs.x << rhs.y;

    return lhs;
}

template <typename T>
std::istream& operator>>(std::istream& lhs, Vector2<T>& rhs)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    lhs >> rhs.x >> rhs.y;

    return lhs;
}

// ---

template <typename T>
Vector2<T> operator+(const Vector2<T>& vec1, const Vector2<T>& vec2)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec1.x + vec2.x, vec1.y + vec2.y};
}

template <typename Vec, typename Val>
Vector2<Vec> operator+(const Vector2<Vec>& vec, const Val& val)
{
    static_assert (std::is_integral<Vec>::value || std::is_floating_point<Vec>::value ||
                   std::is_integral<Val>::value || std::is_floating_point<Val>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec.x + val, vec.y + val};
}

template <typename Val, typename Vec>
Vector2<Vec> operator+(const Val& val, const Vector2<Vec>& vec)
{
    static_assert (std::is_integral<Val>::value || std::is_floating_point<Val>::value ||
                   std::is_integral<Vec>::value || std::is_floating_point<Vec>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {val + vec.x, val + vec.y};
}

// ---

template <typename T>
Vector2<T> operator-(const Vector2<T>& vec1, const Vector2<T>& vec2)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec1.x - vec2.x, vec1.y - vec2.y};
}

template <typename Vec, typename Val>
Vector2<Vec> operator-(const Vector2<Vec>& vec, const Val& val)
{
    static_assert (std::is_integral<Vec>::value || std::is_floating_point<Vec>::value ||
                   std::is_integral<Val>::value || std::is_floating_point<Val>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec.x - val, vec.y - val};
}

template <typename Val, typename Vec>
Vector2<Vec> operator-(const Val& val, const Vector2<Vec>& vec)
{
    static_assert (std::is_integral<Val>::value || std::is_floating_point<Val>::value ||
                   std::is_integral<Vec>::value || std::is_floating_point<Vec>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {val - vec.x, val - vec.y};
}

template <typename T>
Vector2<T> operator-(const Vector2<T>& vec)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {-vec.x, -vec.y};
}

// ---

template <typename T>
Vector2<T> operator*(const Vector2<T>& vec1, const Vector2<T>& vec2)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec1.x * vec2.x, vec1.y * vec2.y};
}

template <typename Vec, typename Val>
Vector2<Vec> operator*(const Vector2<Vec>& vec, const Val& val)
{
    static_assert (std::is_integral<Vec>::value || std::is_floating_point<Vec>::value ||
                   std::is_integral<Val>::value || std::is_floating_point<Val>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec.x * val, vec.y * val};
}

template <typename Val, typename Vec>
Vector2<Vec> operator*(const Val& val, const Vector2<Vec>& vec)
{
    static_assert (std::is_integral<Val>::value || std::is_floating_point<Val>::value ||
                   std::is_integral<Vec>::value || std::is_floating_point<Vec>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {val * vec.x, val + vec.y};
}

// ---

template <typename T>
Vector2<T> operator/(const Vector2<T>& vec1, const Vector2<T>& vec2)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec1.x / vec2.x, vec1.y / vec2.y};
}

template <typename Vec, typename Val>
Vector2<Vec> operator/(const Vector2<Vec>& vec, const Val& val)
{
    static_assert (std::is_integral<Vec>::value || std::is_floating_point<Vec>::value ||
                   std::is_integral<Val>::value || std::is_floating_point<Val>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec.x / val, vec.y / val};
}

template <typename Val, typename Vec>
Vector2<Vec> operator/(const Val& val, const Vector2<Vec> vec)
{
    static_assert (std::is_integral<Val>::value || std::is_floating_point<Val>::value ||
                   std::is_integral<Vec>::value || std::is_floating_point<Vec>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {val / vec.x, val / vec.y};
}

// ****************************************************************************************************************** //

template <typename MT>
class Vector3 {
    static_assert (std::is_integral<MT>::value || std::is_floating_point<MT>::value,
                   "ERROR: template parameter is not an integral or floating point type");
public:
    // --- Vector3: data structures ---

    alignas (16) union {
        struct {
            MT x;
            MT y;
            MT z;
        };
        MT xyz[3];
    };

    // --- Vector3: constructors and deconstructors ---

    Vector3()
    : xyz{0.0, 0.0, 0.0}
    {
    }

    Vector3(const MT& val)
    : xyz{val, val, val}
    {
    }

    Vector3(const MT& x, const MT& y, const MT& z)
    : xyz{x, y, z}
    {
    }

    Vector3(const Vector2<MT>& vec)
    : xyz{vec.x, vec.y, 0.0}
    {
    }

    Vector3(const Vector2<MT>& vec, const MT& val)
    : xyz{vec.x, vec.y, val}
    {
    }

    Vector3(const Vector4<MT>& vec)
    : xyz{vec.x, vec.y, vec.z}
    {
    }

    Vector3(const Vector3& rhs)
    : xyz{rhs.x, rhs.y, rhs.z}
    {
    }

    Vector3(Vector3&& rhs)
    : xyz{std::move(rhs.x), std::move(rhs.y), std::move(rhs.z)}
    {
    }

    virtual ~Vector3()
    {
    }

    // --- Vector3: default operators ---

    Vector3& operator=(const Vector3& rhs)
    {
        if (this != &rhs)
        {
            x = rhs.x;
            y = rhs.y;
            z = rhs.z;
        }

        return *this;
    }

    Vector3& operator=(Vector3&& rhs)
    {
        if (this != &rhs)
        {
            x = std::move(rhs.x);
            y = std::move(rhs.y);
            z = std::move(rhs.z);
        }

        return *this;
    }

    bool operator==(const Vector3& rhs) const
    {
        return x == rhs.x && y == rhs.y && z == rhs.z;
    }

    bool operator!=(const Vector3& rhs) const
    {
        return !(*this == rhs);
    }

    const MT operator[](const size_t& index) const
    {
        return xyz[index];
    }

    MT& operator[](const size_t& index)
    {
        return xyz[index];
    }

    Vector3& operator~()
    {
        x = -x;
        y = -y;
        z = -z;

        return *this;
    }

    Vector3 operator-() const
    {
        return {-x, -y, -z};
    }

    // --- Vector3: other operators ---

    Vector3& operator+=(const Vector3& rhs)
    {
        x += rhs.x;
        y += rhs.y;
        z += rhs.z;

        return *this;
    }

    template <typename T>
    Vector3& operator+=(const T& val)
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        x += val;
        y += val;
        z += val;

        return *this;
    }

    Vector3 operator+(const Vector3& rhs) const
    {
        return {x + rhs.x, y + rhs.y, z + rhs.z};
    }

    template <typename T>
    Vector3 operator+(const T& val) const
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        return {x + val, y + val, z + val};
    }

    // ---

    Vector3& operator-=(const Vector3& rhs)
    {
        x -= rhs.x;
        y -= rhs.y;
        z -= rhs.z;

        return *this;
    }

    template <typename T>
    Vector3& operator-=(const T& val)
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        x -= val;
        y -= val;
        z -= val;

        return *this;
    }

    Vector3 operator-(const Vector3& rhs) const
    {
        return {x - rhs.x, y - rhs.y, z - rhs.z};
    }

    template <typename T>
    Vector3 operator-(const T& val) const
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        return {x - val, y - val, z - val};
    }

    // ---

    Vector3& operator*=(const Vector3& rhs)
    {
        x *= rhs.x;
        y *= rhs.y;
        z *= rhs.z;

        return *this;
    }

    template <typename T>
    Vector3& operator*=(const T& val)
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        x *= val;
        y *= val;
        z *= val;

        return *this;
    }

    Vector3 operator*(const Vector3& rhs) const
    {
        return {x * rhs.x, y * rhs.y, z * rhs.z};
    }

    template <typename T>
    Vector3 operator*(const T& val) const
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        return {x * val, y * val, z * val};
    }

    // ---

    Vector3& operator/=(const Vector3& rhs)
    {
        x /= rhs.x;
        y /= rhs.y;
        z /= rhs.z;

        return *this;
    }

    template <typename T>
    Vector3& operator/=(const T& val)
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        x /= val;
        y /= val;
        z /= val;

        return *this;
    }

    Vector3 operator/(const Vector3& rhs) const
    {
        return {x / rhs.x, y / rhs.y, z / rhs.z};
    }

    template <typename T>
    Vector3 operator/(const T& val) const
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        return {x / val, y / val, z / val};
    }

    // --- Vector3: public methods ---

    Vector3 crossProduct(const Vector3& vec) const
    {
        return {y * vec.z - z * vec.y, z * vec.x - x * vec.z, x * vec.y - y * vec.x};
    }

    MT distanceToLine(const Vector3& origin, const Vector3& direction) const
    {
         if (direction.isNull())
            return (*this - origin).length();

        return (*this - (origin + dotProduct(*this - origin, direction) * direction)).length();
    }

    MT distanceToLinePoints(const Vector3& point1, const Vector3& point2) const
    {
        const Vector3 direction = point2 - point1;

        if (direction.isNull())
            return (*this - point1).length();

        return (*this - (point1 + dotProduct(*this - point1, direction) * direction)).length();
    }

    MT distanceToPlane(const Vector3& origin, const Vector3& normal) const
    {
        return dotProduct(*this - origin, normal);
    }

    MT distanceToPlane(const Vector3& point1, const Vector3& point2, const Vector3& point3) const
    {
        return dotProduct(*this - point1, Vector3::normal(point2 - point1, point3 - point1));
    }

    MT distanceToPoint(const Vector3& point) const
    {
        return (*this - point).length();
    }

    MT dotProduct(const Vector3& vec) const
    {
        return x * vec.x + y * vec.y + z * vec.z;
    }

    bool isNull() const
    {
        return x == 0.0 && y == 0.0 && z == 0.0;
    }

    MT length() const
    {
        return std::sqrt(x * x + y * y + z * z);
    }

    MT lengthSquared() const
    {
        return x * x + y * y + z * z;
    }

    Vector3 normal(const Vector3& vec) const
    {
        return crossProduct(vec).normalized();
    }

    Vector3 normal(const Vector3& point2, const Vector3& point3) const
    {
        return crossProduct(point2 - *this, point3 - *this).normalized();
    }

    void normalize()
    {
        *this /= std::sqrt(x * x + y * y + z * z);
    }

    Vector3 normalized() const
    {
        const MT len = std::sqrt(x * x + y * y + z * z);

        return {x / len, y / len, z / len};
    }

    std::string string() const
    {
        std::stringstream result;

        result << x << ", " << y << ", " << z;

        return result.str();
    }

    Vector2<MT> vector2() const
    {
        return {x, y};
    }

    Vector4<MT> vector4() const
    {
        return {x, y, z, 0.0};
    }

    // --- Vector3: static methods ---

    static Vector3 toCrossProduct(const Vector3& vec1, const Vector3& vec2)
    {
        return {vec1.y * vec2.z - vec1.z * vec2.y,
                vec1.z * vec2.x - vec1.x * vec2.z,
                vec1.x * vec2.y - vec1.y * vec2.x};
    }

    static MT toDotProduct(const Vector3& vec1, const Vector3& vec2)
    {
        return vec1.x * vec2.x + vec1.y * vec2.y + vec1.z * vec2.z;
    }

    static Vector3 toNormal(const Vector3& vec1, const Vector3& vec2)
    {
        return toCrossProduct(vec1, vec2).normalized();
    }

    static Vector3 toNormal(const Vector3& point1, const Vector3& point2, const Vector3& point3)
    {
        return toCrossProduct(point2 - point1, point3 - point1).normalized();
    }

    static std::string toString(const Vector3& vec)
    {
        std::stringstream result;

        result << vec.x << ", " << vec.y << ", " << vec.z;

        return result.str();
    }

    static Vector2<MT> toVector2(const Vector3& vec)
    {
        return {vec.x, vec.y};
    }

    static Vector4<MT> toVector4(const Vector3& vec)
    {
        return {vec.x, vec.y, vec.z, 0.0};
    }
};

// --- Vector3: global operators ---

template <typename T>
bool operator==(const Vector3<T>& lhs, const Vector3<T>& rhs)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return lhs.x == rhs.x && lhs.y == rhs.y && lhs.z == rhs.z;
}

template <typename T>
bool operator!=(const Vector3<T> lhs, const Vector3<T>& rhs)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return !(lhs == rhs);
}

template <typename T>
std::ostream& operator<<(std::ostream& lhs, const Vector3<T>& rhs)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    lhs << rhs.x << rhs.y << rhs.z;

    return lhs;
}

template <typename T>
std::istream& operator>>(std::istream& lhs, Vector3<T>& rhs)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    lhs >> rhs.x >> rhs.y >> rhs.z;

    return lhs;
}

// ---

template <typename T>
Vector3<T> operator+(const Vector3<T>& vec1, const Vector3<T>& vec2)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec1.x + vec2.x, vec1.y + vec2.y, vec1.z + vec2.z};
}

template <typename Vec, typename Val>
Vector3<Vec> operator+(const Vector3<Vec>& vec, const Val& val)
{
    static_assert (std::is_integral<Vec>::value || std::is_floating_point<Vec>::value ||
                   std::is_integral<Val>::value || std::is_floating_point<Val>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec.x + val, vec.y + val, vec.z + val};
}

template <typename Val, typename Vec>
Vector3<Vec> operator+(const Val& val, const Vector3<Vec>& vec)
{
    static_assert (std::is_integral<Val>::value || std::is_floating_point<Val>::value ||
                   std::is_integral<Vec>::value || std::is_floating_point<Vec>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {val + vec.x, val + vec.y, val + vec.z};
}

// ---

template <typename T>
Vector3<T> operator-(const Vector3<T>& vec1, const Vector3<T>& vec2)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec1.x - vec2.x, vec1.y - vec2.y, vec1.z - vec2.z};
}

template <typename Vec, typename Val>
Vector3<Vec> operator-(const Vector3<Vec>& vec, const Val& val)
{
    static_assert (std::is_integral<Vec>::value || std::is_floating_point<Vec>::value ||
                   std::is_integral<Val>::value || std::is_floating_point<Val>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec.x - val, vec.y - val, vec.z - val};
}

template <typename Val, typename Vec>
Vector3<Vec> operator-(const Val& val, const Vector3<Vec>& vec)
{
    static_assert (std::is_integral<Val>::value || std::is_floating_point<Val>::value ||
                   std::is_integral<Vec>::value || std::is_floating_point<Vec>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {val - vec.x, val - vec.y, val - vec.z};
}

template <typename T>
Vector3<T> operator-(const Vector3<T>& vec)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {-vec.x, -vec.y, -vec.z};
}

// ---

template <typename T>
Vector3<T> operator*(const Vector3<T>& vec1, const Vector3<T>& vec2)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec1.x * vec2.x, vec1.y * vec2.y, vec1.z * vec2.z};
}

template <typename Vec, typename Val>
Vector3<Vec> operator*(const Vector3<Vec>& vec, const Val& val)
{
    static_assert (std::is_integral<Vec>::value || std::is_floating_point<Vec>::value ||
                   std::is_integral<Val>::value || std::is_floating_point<Val>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec.x * val, vec.y * val, vec.z * val};
}

template <typename Val, typename Vec>
Vector3<Vec> operator*(const Val& val, const Vector3<Vec>& vec)
{
    static_assert (std::is_integral<Val>::value || std::is_floating_point<Val>::value ||
                   std::is_integral<Vec>::value || std::is_floating_point<Vec>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {val * vec.x, val * vec.y, val * vec.z};
}

// ---

template <typename T>
Vector3<T> operator/(const Vector3<T>& vec1, const Vector3<T>& vec2)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec1.x / vec2.x, vec1.y / vec2.y, vec1.z / vec2.z};
}

template <typename Vec, typename Val>
Vector3<Vec> operator/(const Vector3<Vec>& vec, const Val& val)
{
    static_assert (std::is_integral<Vec>::value || std::is_floating_point<Vec>::value ||
                   std::is_integral<Val>::value || std::is_floating_point<Val>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec.x / val, vec.y / val, vec.z / val};
}

template <typename Val, typename Vec>
Vector3<Vec> operator/(const Val& val, const Vector3<Vec> vec)
{
    static_assert (std::is_integral<Val>::value || std::is_floating_point<Val>::value ||
                   std::is_integral<Vec>::value || std::is_floating_point<Vec>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {val / vec.x, val / vec.y, val / vec.z};
}

// ****************************************************************************************************************** //

template <typename MT>
class Vector4 {
    static_assert (std::is_integral<MT>::value || std::is_floating_point<MT>::value,
                   "ERROR: template parameter is not an integral or floating point type");
public:
    // --- Vector4: data structures ---

    alignas (16) union {
        struct {
            MT x;
            MT y;
            MT z;
            MT w;
        };
        MT xyzw[4];
    };

    // --- Vector4: constructors and deconstructors ---

    Vector4()
    : xyzw{0.0, 0.0, 0.0, 0.0}
    {
    }

    Vector4(const MT& val)
    : xyzw{val, val, val, val}
    {
    }

    Vector4(const MT& x, const MT& y, const MT& z, const MT& w)
    : xyzw{x, y, z, w}
    {
    }

    Vector4(const Vector2<MT>& vec)
    : xyzw{vec.x, vec.y, 0.0, 0.0}
    {
    }

    Vector4(const Vector2<MT>& vec, const MT& val1, const MT& val2)
    : xyzw{vec.x, vec.y, val1, val2}
    {
    }

    Vector4(const Vector2<MT>& vec1, const Vector2<MT>& vec2)
    : xyzw{vec1.x, vec1.y, vec2.x, vec2.y}
    {
    }

    Vector4(const Vector3<MT>& vec)
    : xyzw{vec.x, vec.y, vec.z, 0.0}
    {
    }

    Vector4(const Vector3<MT>& vec, const MT& val)
    : xyzw{vec.x, vec.y, vec.z, val}
    {
    }

    Vector4(const Vector4& rhs)
    : xyzw{rhs.x, rhs.y, rhs.z, rhs.w}
    {
    }

    Vector4(Vector4&& rhs)
    : xyzw{std::move(rhs.x), std::move(rhs.y), std::move(rhs.z), std::move(rhs.w)}
    {
    }

    virtual ~Vector4()
    {
    }

    // --- Vector4: default operators ---

    Vector4& operator=(const Vector4& rhs)
    {
        if (this != &rhs)
        {
            x = rhs.x;
            y = rhs.y;
            z = rhs.z;
            w = rhs.w;
        }

        return *this;
    }

    Vector4& operator=(Vector4&& rhs)
    {
        if (this != &rhs)
        {
            x = std::move(rhs.x);
            y = std::move(rhs.y);
            z = std::move(rhs.z);
            w = std::move(rhs.w);
        }

        return *this;
    }

    bool operator==(const Vector4& rhs) const
    {
        return x == rhs.x && y == rhs.y && z == rhs.z && w == rhs.w;
    }

    bool operator!=(const Vector4& rhs) const
    {
        return !(*this == rhs);
    }

    const MT operator[](const size_t& index) const
    {
        return xyzw[index];
    }

    MT& operator[](const size_t& index)
    {
        return xyzw[index];
    }

    Vector4& operator~()
    {
        x = -x;
        y = -y;
        z = -z;
        w = -w;

        return *this;
    }

    Vector4 operator-() const
    {
        return {-x, -y, -z, -w};
    }

    // --- Vector4: other operators ---

    Vector4& operator+=(const Vector4& rhs)
    {
        x += rhs.x;
        y += rhs.y;
        z += rhs.z;

        return *this;
    }

    template <typename T>
    Vector4& operator+=(const T& val)
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        x += val;
        y += val;
        z += val;
        w += val;

        return *this;
    }

    Vector4 operator+(const Vector4& rhs) const
    {
        return {x + rhs.x, y + rhs.y, z + rhs.z, w + rhs.w};
    }

    template <typename T>
    Vector4 operator+(const T& val) const
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        return {x + val, y + val, z + val, w + val};
    }

    // ---

    Vector4& operator-=(const Vector4& rhs)
    {
        x -= rhs.x;
        y -= rhs.y;
        z -= rhs.z;
        w -= rhs.w;

        return *this;
    }

    template <typename T>
    Vector4& operator-=(const T& val)
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        x -= val;
        y -= val;
        z -= val;
        w -= val;

        return *this;
    }

    Vector4 operator-(const Vector4& rhs) const
    {
        return {x - rhs.x, y - rhs.y, z - rhs.z, w - rhs.w};
    }

    template <typename T>
    Vector4 operator-(const T& val) const
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        return {x - val, y - val, z - val, w - val};

    }

    // ---

    Vector4& operator*=(const Vector4& rhs)
    {
        x *= rhs.x;
        y *= rhs.y;
        z *= rhs.z;
        w *= rhs.w;

        return *this;
    }

    template <typename T>
    Vector4& operator*=(const T& val)
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        x *= val;
        y *= val;
        z *= val;
        w *= val;

        return *this;
    }

    Vector4 operator*(const Vector4& rhs) const
    {
        return {x * rhs.x, y * rhs.y, z * rhs.z, w * rhs.w};
    }

    template <typename T>
    Vector4 operator*(const T& val) const
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        return {x * val, y * val, z * val, w * val};
    }

    // ---

    Vector4& operator/=(const Vector4& rhs)
    {
        x /= rhs.x;
        y /= rhs.y;
        z /= rhs.z;
        w /= rhs.w;

        return *this;
    }

    template <typename T>
    Vector4& operator/=(const T& val)
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        x /= val;
        y /= val;
        z /= val;
        w /= val;

        return *this;
    }

    Vector4 operator/(const Vector4& rhs) const
    {
        return {x / rhs.x, y / rhs.y, z / rhs.z, w / rhs.w};
    }

    template <typename T>
    Vector4 operator/(const T& val) const
    {
        static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                       "ERROR: template parameter is not an integral or floating point type");

        return {x / val, y / val, z / val, w / val};
    }

    // --- Vector4: public methods ---

    MT dotProduct(const Vector4& vec) const
    {
        return x * vec.x + y * vec.y + z * vec.z + w * vec.w;
    }

    bool isNull() const
    {
        return x == 0.0 && y == 0.0 && z == 0.0 && w == 0.0;
    }

    MT length() const
    {
        return std::sqrt(x * x + y * y + z * z + w * w);
    }

    MT lengthSquared() const
    {
        return x * x + y * y + z * z + w * w;
    }

    void normalize()
    {
        *this /= std::sqrt(x * x + y * y + z * z + w * w);
    }

    Vector4 normalized() const
    {
        const MT len = std::sqrt(x * x + y * y + z * z + w * w);

        return {x / len, y / len, z / len, w / len};
    }

    std::string string() const
    {
        std::stringstream result;

        result << x << ", " << y << ", " << z << ", " << w;

        return result.str();
    }

    Vector2<MT> vector2() const
    {
        return {x, y};
    }

    Vector3<MT> vector3() const
    {
        return {x, y, z};
    }

    // --- Vector4: static methods ---

    static MT toDotProduct(const Vector4& vec1, const Vector4& vec2)
    {
        return vec1.x * vec2.x + vec1.y * vec2.y + vec1.z * vec2.z + vec1.w * vec2.w;
    }

    static std::string toString(const Vector4& vec)
    {
        std::stringstream result;

        result << vec.x << ", " << vec.y << ", " << vec.z << ", " << vec.w;

        return result.str();
    }

    static Vector2<MT> toVector2(const Vector4& vec)
    {
        return {vec.x, vec.y};
    }

    static Vector3<MT> toVector3(const Vector4& vec)
    {
        return {vec.x, vec.x, vec.z};
    }
};

// --- Vector4: global operators ---

template <typename T>
bool operator==(const Vector4<T>& lhs, const Vector4<T>& rhs)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return lhs.x == rhs.x && lhs.y == rhs.y && lhs.z == rhs.z && lhs.w == rhs.w;
}

template <typename T>
bool operator!=(const Vector4<T> lhs, const Vector4<T>& rhs)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return !(lhs == rhs);
}

template <typename T>
std::ostream& operator<<(std::ostream& lhs, const Vector4<T>& rhs)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    lhs << rhs.x << rhs.y << rhs.z << rhs.w;

    return lhs;
}

template <typename T>
std::istream& operator>>(std::istream& lhs, Vector4<T>& rhs)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    lhs >> rhs.x >> rhs.y >> rhs.z >> rhs.w;

    return lhs;
}

// ---

template <typename T>
Vector4<T> operator+(const Vector4<T>& vec1, const Vector4<T>& vec2)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec1.x + vec2.x, vec1.y + vec2.y, vec1.z + vec2.z, vec1.w + vec2.w};
}

template <typename Vec, typename Val>
Vector4<Vec> operator+(const Vector4<Vec>& vec, const Val& val)
{
    static_assert (std::is_integral<Vec>::value || std::is_floating_point<Vec>::value ||
                   std::is_integral<Val>::value || std::is_floating_point<Val>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec.x + val, vec.y + val, vec.z + val, vec.w + val};
}

template <typename Val, typename Vec>
Vector4<Vec> operator+(const Val& val, const Vector4<Vec>& vec)
{
    static_assert (std::is_integral<Val>::value || std::is_floating_point<Val>::value ||
                   std::is_integral<Vec>::value || std::is_floating_point<Vec>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {val + vec.x, val + vec.y, val + vec.z, val + vec.w};
}

// ---

template <typename T>
Vector4<T> operator-(const Vector4<T>& vec1, const Vector4<T>& vec2)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec1.x - vec2.x, vec1.y - vec2.y, vec1.z - vec2.z, vec1.w - vec2.w};
}

template <typename Vec, typename Val>
Vector4<Vec> operator-(const Vector4<Vec>& vec, const Val& val)
{
    static_assert (std::is_integral<Vec>::value || std::is_floating_point<Vec>::value ||
                   std::is_integral<Val>::value || std::is_floating_point<Val>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec.x - val, vec.y - val, vec.z - val, vec.w - val};
}

template <typename Val, typename Vec>
Vector4<Vec> operator-(const Val& val, const Vector4<Vec>& vec)
{
    static_assert (std::is_integral<Val>::value || std::is_floating_point<Val>::value ||
                   std::is_integral<Vec>::value || std::is_floating_point<Vec>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {val - vec.x, val - vec.y, val - vec.z, val - vec.w};
}

template <typename T>
Vector4<T> operator-(const Vector4<T>& vec)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {-vec.x, -vec.y, -vec.z, -vec.w};
}

// ---

template <typename T>
Vector4<T> operator*(const Vector4<T>& vec1, const Vector4<T>& vec2)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec1.x * vec2.x, vec1.y * vec2.y, vec1.z * vec2.z, vec1.w * vec2.w};
}

template <typename Vec, typename Val>
Vector4<Vec> operator*(const Vector4<Vec>& vec, const Val& val)
{
    static_assert (std::is_integral<Vec>::value || std::is_floating_point<Vec>::value ||
                   std::is_integral<Val>::value || std::is_floating_point<Val>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec.x * val, vec.y * val, vec.z * val, vec.w * val};
}

template <typename Val, typename Vec>
Vector4<Vec> operator*(const Val& val, const Vector4<Vec>& vec)
{
    static_assert (std::is_integral<Val>::value || std::is_floating_point<Val>::value ||
                   std::is_integral<Vec>::value || std::is_floating_point<Vec>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {val * vec.x, val + vec.y, val * vec.z, val * vec.w};
}

// ---

template <typename T>
Vector4<T> operator/(const Vector4<T>& vec1, const Vector4<T>& vec2)
{
    static_assert (std::is_integral<T>::value || std::is_floating_point<T>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec1.x / vec2.x, vec1.y / vec2.y, vec1.z / vec2.z, vec1.w / vec2.w};
}

template <typename Vec, typename Val>
Vector4<Vec> operator/(const Vector4<Vec>& vec, const Val& val)
{
    static_assert (std::is_integral<Vec>::value || std::is_floating_point<Vec>::value ||
                   std::is_integral<Val>::value || std::is_floating_point<Val>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {vec.x / val, vec.y / val, vec.z / val, vec.w / val};
}

template <typename Val, typename Vec>
Vector4<Vec> operator/(const Val& val, const Vector4<Vec> vec)
{
    static_assert (std::is_integral<Val>::value || std::is_floating_point<Val>::value ||
                   std::is_integral<Vec>::value || std::is_floating_point<Vec>::value,
                   "ERROR: template parameter is not an integral or floating point type");

    return {val / vec.x, val / vec.y, val / vec.z, val / vec.w};
}
