#include "filesystem/NCFile.hxx"

/*
 * qCompress()/qUncompress() does not behave like the real zlib compress2/uncompress functions,
 * the Qt functions add a quint32 at the beginning of the QByteArray holding the original size
 */

// --- internal constants and stuff ---

const quint8 All = NC::FileData::HSize | NC::FileData::Position | NC::FileData::CSize | NC::FileData::OSize
                 | NC::FileData::NameLength | NC::FileData::Name | NC::FileData::CData | NC::FileData::OData;

// --- constructors and deconstructors ---

NCFile::NCFile(QObject *parent,
               const quint32& hsize, const quint32& pos, const quint32& csize, const quint32& osize,
               const quint32& nlen, const QString& name, const QByteArray& cdata)
: QObject(parent), _hsize(hsize), _position(pos), _csize(csize), _osize(osize), _name_length(nlen), _name(name),
  _cdata(cdata)
{
    emit sigFileData(this, All);
}

NCFile::NCFile(const NCFile& rhs)
: QObject(rhs.parent()), _hsize(rhs._hsize), _position(rhs._position), _csize(rhs._csize), _osize(rhs._osize),
  _name_length(rhs._name_length), _name(rhs._name), _cdata(rhs._cdata)
{
    emit sigFileData(this, All);
}

NCFile::NCFile(const NCFile&& rhs)
: QObject(std::move(rhs.parent())), _hsize(std::move(rhs._hsize)), _position(std::move(rhs._position)),
  _csize(std::move(rhs._csize)), _osize(std::move(rhs._osize)), _name_length(std::move(rhs._name_length)),
  _name(std::move(rhs._name)), _cdata(std::move(rhs._cdata))
{
    emit sigFileData(this, All);
}

NCFile::~NCFile()
{
}

// --- operators ---

NCFile& NCFile::operator=(const NCFile& rhs)
{
    if (this != &rhs)
    {
        _hsize = rhs._hsize;
        _position = rhs._position;
        _csize = rhs._csize;
        _osize = rhs._osize;
        _name_length = rhs._name_length;
        _name = rhs._name;
        _cdata = rhs._cdata;
    }

    return *this;
}

NCFile& NCFile::operator=(const NCFile&& rhs)
{
    if (this != &rhs)
    {
        _hsize = std::move(rhs._hsize);
        _position = std::move(rhs._position);
        _csize = std::move(rhs._csize);
        _osize = std::move(rhs._osize);
        _name_length = std::move(rhs._name_length);
        _name = std::move(rhs._name);
        _cdata = std::move(rhs._cdata);
    }

    return *this;
}

bool NCFile::operator==(const NCFile& rhs) const
{
    return _hsize == rhs._hsize
           && _position == rhs._position
           && _csize == rhs._csize
           && _osize == rhs._osize
           && _name_length == rhs._name_length
           && _name == rhs._name
           && _cdata == rhs._cdata;
}

bool NCFile::operator!=(const NCFile& rhs) const
{
    return !(*this == rhs);
}

// --- setters and getters ---

void NCFile::setHSize(const quint32& size)
{
    _hsize = size;
}

quint32 NCFile::hSize() const
{
    return _hsize;
}

void NCFile::setPosition(const quint32& pos)
{
    _position = pos;
}

quint32 NCFile::position() const
{
    return _position;
}

void NCFile::setCSize(const quint32& size)
{
    _csize = size;
}

quint32 NCFile::cSize() const
{
    return _csize;
}

void NCFile::setOSize(const quint32& size)
{
    _osize = size;
}

quint32 NCFile::oSize() const
{
    return _osize;
}

void NCFile::setNameLength(const quint32& len)
{
    _name_length = len;
}

quint32 NCFile::nameLength() const
{
    return _name_length;
}

void NCFile::setName(const QString& name)
{
    _name = name;
}

QString NCFile::name() const
{
    return _name;
}

void NCFile::setCData(const QByteArray& data)
{
    _cdata = data;
}

QByteArray NCFile::cData() const
{
    return _cdata;
}

void NCFile::setOData(const QByteArray& data)
{
    _cdata = qCompress(data, 9).mid(sizeof (_csize));
}

QByteArray NCFile::oData() const
{
    // this is the only way it works, qCompress/qDecompress is broken!
    QByteArray data(sizeof (_osize) + _cdata.size(), '\0');

    data.replace(sizeof (_osize),_csize, _cdata);

    return qUncompress(data);
}

// --- status methods ---

quint32 NCFile::size() const
{
    return HSize + _name_length + _cdata.size();
}

quint32 NCFile::headerSize() const
{
    return HSize;
}

bool NCFile::valid() const
{
    return !(_hsize == Error || _position == Error || _csize == Error || _osize == Error || _name_length == Error
           || _name.isEmpty() || _cdata.isEmpty());
}
